/**
 * A `Font` is used in conjunction with the `Text` primitive to render text, such as axis labels.
 *
 * Font image and data files can be generated with `msdf-bmfont`:
 * ```sh
 * $ npm i -g msdf-bmfont-xml
 * $ msdf-bmfont -f json -s 24 -t sdf --smart-size Lato-Regular.ttf
 * ```
 */
export class Font {
    /**
     *
     * @param cg
     * @param image The image produced by msdf-bmfont.
     * @param json The json object produced by msdf-bmfont.
     */
    constructor(cg, image, json) {
        /** @internal */
        this.glyphs = [];
        this.texture = cg.regl.texture({
            data: image,
            mag: "linear",
            min: "linear",
            flipY: true,
        });
        const scale = 1.0 / json.info.size;
        this.lineHeight = scale * json.common.lineHeight;
        const { scaleW, scaleH } = json.common;
        for (const char of json.chars) {
            this.glyphs[char.id] = {
                id: char.id,
                xoffset: scale * char.xoffset,
                yoffset: scale * char.yoffset,
                width: scale * char.width,
                height: scale * char.height,
                xadvance: scale * char.xadvance,
                // Original image is flipped, and the original coordinates are from
                // top left to bottom right, so we need to flip the y-coordinates (1 - y)
                // AND swap their order (y0, y1 = y1, y0) to make things line up normally
                // in webgl-land.
                uv: [
                    char.x / scaleW,
                    1.0 - (char.y + char.height) / scaleH,
                    (char.x + char.width) / scaleW,
                    1.0 - char.y / scaleH,
                ],
            };
        }
        // Create the kerning table for quick lookups.
        this.maxid = this.glyphs.length;
        this.kernTable = new Int8Array(this.maxid * this.maxid);
        for (const kern of json.kernings) {
            this.kernTable[kern.first * this.maxid + kern.second] = scale * kern.amount;
        }
    }
    /** @internal */
    kern(first, second) {
        return this.kernTable[first * this.maxid + second];
    }
    /** Release all GPU resources and render this Font instance unusable. */
    dispose() {
        this.texture.destroy();
    }
}
//# sourceMappingURL=font.js.map