import { vec2 } from "gl-matrix";
import { Composite } from "./composite";
import { Text } from "../primitives/text";
import { VLines } from "../primitives/vlines";
import { HLines } from "../primitives/hlines";
import { LineSegments } from "../primitives/line-segments";
const DEFAULTS = {
    axisColor: [0, 0, 0, 1],
    axisWidth: 1,
    labelAngle: 0,
    labelColor: [0, 0, 0, 1],
    labelPad: 3,
    labelSide: -1,
    labelSize: 12,
    tickColor: [0, 0, 0, 1],
    tickLength: 12,
    tickOffset: 0,
    tickWidth: 1,
    minorTickColor: [0, 0, 0, 1],
    minorTickLength: 6,
    minorTickOffset: 0,
    minorTickWidth: 1,
    minorTicks: [],
};
export class Axis extends Composite {
    /**
     *
     * @param start One endpoint of the axis.
     * @param end The other endpoint of the axis.
     * @param ticks The position of the ticks as a distance from `start`.
     * @param labels The string labels for the ticks.
     * @param font The `Font` instance used to render the tick labels.
     */
    constructor(cg, coords, start, end, ticks, labels, font, options = {}) {
        super();
        this.axis = [];
        this.texts = [];
        this.ticks = [];
        this.minorTicks = [];
        const opts = Object.assign(Object.assign({}, DEFAULTS), options);
        const { axisWidth, axisColor, labelAngle, labelPad, labelSide, labelSize, labelColor, labelAnchor, tickOffset, tickLength, tickWidth, tickColor, minorTickOffset, minorTickLength, minorTickWidth, minorTickColor, } = opts;
        this.texts = [];
        const tickPoints = [];
        const axisPoints = [...start, ...end];
        const p0p1 = vec2.sub(vec2.create(), end, start);
        const dirWorld = vec2.normalize(vec2.create(), p0p1);
        const startScreen = coords.toRange(start);
        const endScreen = coords.toRange(end);
        const p0p1Screen = vec2.sub(vec2.create(), endScreen, startScreen);
        const orthoScreen = vec2.normalize(vec2.create(), vec2.fromValues(-p0p1Screen[1], p0p1Screen[0]));
        let anchor = labelAnchor;
        if (!anchor) {
            anchor = vec2.scale(vec2.create(), orthoScreen, labelSide);
            const qAnchor = 1 / Math.max(Math.abs(anchor[0]), Math.abs(anchor[1]));
            vec2.scale(anchor, anchor, qAnchor);
        }
        const labelOpts = {
            align: 0.5,
            anchor: anchor,
            angle: labelAngle,
            color: labelColor,
            size: labelSize,
        };
        for (let tickIndex = 0; tickIndex < ticks.length; tickIndex++) {
            const tick = ticks[tickIndex];
            const label = labels[tickIndex];
            // tickCenter = p0 + dt * direction
            const tickCenterScreen = coords.toRange(vec2.add(vec2.create(), start, vec2.scale(vec2.create(), dirWorld, tick)));
            // tickCenterScreen = tickCenterScreen + tickOffset * orthoScreen
            vec2.add(tickCenterScreen, tickCenterScreen, vec2.scale(vec2.create(), orthoScreen, tickOffset));
            // halfTick = 0.5 * tickLength * orthoScreen
            const halfTick = vec2.scale(vec2.create(), orthoScreen, 0.5 * tickLength);
            // hi = tickCenterScreen + halfTick
            const hi = vec2.add(vec2.create(), tickCenterScreen, halfTick);
            // lo = tickCenterScreen - halfTick
            const lo = vec2.sub(vec2.create(), tickCenterScreen, halfTick);
            // Push the tick line.
            tickPoints.push(...coords.toDomain(lo), ...coords.toDomain(hi));
            // labelPosition = tickCenterScreen + labelSide * (0.5 * tickLength + labelPad) * orthoScreen
            const labelPosition = vec2.add(vec2.create(), tickCenterScreen, vec2.scale(vec2.create(), orthoScreen, -labelSide * (0.5 * tickLength + labelPad)));
            this.texts.push(new Text(cg, font, label, coords.toDomain(labelPosition), labelOpts));
        }
        // Minor ticks.
        const minorTickPoints = [];
        for (let tickIndex = 0; tickIndex < opts.minorTicks.length; tickIndex++) {
            const tick = opts.minorTicks[tickIndex];
            // tickCenter = p0 + dt * direction
            const tickCenterScreen = coords.toRange(vec2.add(vec2.create(), start, vec2.scale(vec2.create(), dirWorld, tick)));
            // tickCenterScreen = tickCenterScreen + tickOffset * orthoScreen
            vec2.add(tickCenterScreen, tickCenterScreen, vec2.scale(vec2.create(), orthoScreen, minorTickOffset));
            // halfTick = 0.5 * tickLength * orthoScreen
            const halfTick = vec2.scale(vec2.create(), orthoScreen, 0.5 * minorTickLength);
            // hi = tickCenterScreen + halfTick
            const hi = vec2.add(vec2.create(), tickCenterScreen, halfTick);
            // lo = tickCenterScreen - halfTick
            const lo = vec2.sub(vec2.create(), tickCenterScreen, halfTick);
            // Push the tick line.
            minorTickPoints.push(...coords.toDomain(lo), ...coords.toDomain(hi));
        }
        const verticalAxis = start[0] === end[0];
        const horizontalAxis = start[1] === end[1];
        if (axisPoints.length > 0) {
            if (verticalAxis) {
                this.axis = new VLines(cg, segmentsToVlines(axisPoints), {
                    widths: axisWidth,
                    colors: axisColor,
                });
            }
            else if (horizontalAxis) {
                this.axis = new HLines(cg, segmentsToHlines(axisPoints), {
                    widths: axisWidth,
                    colors: axisColor,
                });
            }
            else {
                this.axis = new LineSegments(cg, axisPoints, {
                    widths: axisWidth,
                    colors: axisColor,
                });
            }
        }
        if (tickPoints.length > 0) {
            if (verticalAxis) {
                this.ticks = new HLines(cg, segmentsToHlines(tickPoints), {
                    widths: tickWidth,
                    colors: tickColor,
                });
            }
            else if (horizontalAxis) {
                this.ticks = new VLines(cg, segmentsToVlines(tickPoints), {
                    widths: tickWidth,
                    colors: tickColor,
                });
            }
            else {
                this.ticks = new LineSegments(cg, tickPoints, {
                    widths: tickWidth,
                    colors: tickColor,
                });
            }
        }
        if (minorTickPoints.length > 0) {
            if (verticalAxis) {
                this.minorTicks = new HLines(cg, segmentsToHlines(minorTickPoints), {
                    widths: minorTickWidth,
                    colors: minorTickColor,
                });
            }
            else if (horizontalAxis) {
                this.minorTicks = new VLines(cg, segmentsToVlines(minorTickPoints), {
                    widths: minorTickWidth,
                    colors: minorTickColor,
                });
            }
            else {
                this.minorTicks = new LineSegments(cg, minorTickPoints, {
                    widths: tickWidth,
                    colors: tickColor,
                });
            }
        }
    }
    /** @internal */
    children() {
        return [this.axis, this.ticks, this.minorTicks, this.texts];
    }
}
function segmentsToHlines(segments) {
    const hlines = [];
    for (let i = 0; i < segments.length / 4; i++) {
        hlines.push(segments[4 * i + 0], segments[4 * i + 2], segments[4 * i + 1]);
    }
    return hlines;
}
function segmentsToVlines(segments) {
    const vlines = [];
    for (let i = 0; i < segments.length / 4; i++) {
        vlines.push(segments[4 * i + 0], segments[4 * i + 1], segments[4 * i + 3]);
    }
    return vlines;
}
//# sourceMappingURL=axis.js.map