function isDataset(obj) {
    return obj.constructor === Dataset;
}
export function createDataset(cg, data) {
    if (isDataset(data)) {
        return data;
    }
    return new Dataset(cg, data);
}
/**
 * `Datasets` are used to store and reference data on the GPU. This is useful
 * for making some operations more efficient, such as animation or switching
 * between `CoordinateSystems`.
 *
 * @example
 * ```ts
 * // Create a reusable DataSet.
 * const data = new Dataset(cg, [0, 0, 1, 1]);
 *
 * // Render a line segment to the viewport.
 * cg.render(coords, viewport, [
 *   new LineSegment(cg, data, {colors: [1, 0, 0, 1], widths: 4]}), // red, width 4
 *   new LineSegment(cg, data, {colors: [0, 0, 1, 1], widths: 1]}), // blue, width 1
 * ]);
 * ```
 */
export class Dataset {
    constructor(cg, data) {
        this.length = 0;
        this.disposed = false;
        /** Update this Dataset with new data, in-place. */
        this.update = (data) => {
            if (this.disposed) {
                throw new Error("This DataSet cannot be updated, it's already been disposed.");
            }
            if (typeof data === "number") {
                this.buffer([data]);
                this.length = 1;
            }
            else if (ArrayBuffer.isView(data)) {
                this.buffer(data);
                this.length = data.length;
            }
            else if (Array.isArray(data)) {
                this.buffer(data);
                if (typeof data[0] === "number") {
                    this.length = data.length;
                }
                else {
                    this.length = data.length * data[0].length;
                }
            }
            return this;
        };
        this.buffer = cg.regl.buffer(1);
        this.update(data);
    }
    /** Releases all GPU resources and renders this instance unusable. */
    dispose() {
        this.buffer.destroy();
        this.disposed = true;
    }
    /** @internal */
    count(size) {
        this.assertFits(size);
        return this.length / size;
    }
    /** @internal */
    divisor(instances, size) {
        const count = this.count(size);
        if (count === 1) {
            return instances;
        }
        return 1;
    }
    assertFits(size) {
        if (this.length % size !== 0) {
            throw new Error(`Attempted to use incompatible size ${size} with data of length ${this.length}.`);
        }
    }
}
//# sourceMappingURL=dataset.js.map